# Copyright 2016, 2017, 2018 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt


"""
netq - Query data across all nodes in fabric

Usage:
   netq help [<text-keywords>]
   netq help verbose
   netq help list

Options:
   all                          : Information about all sub components
   around                       : Go back in time to around ...
   between                      : Specify start and end times for output desired
   config                       : Configure NetQ
   count                        : Count of matching entries
   example                      : Show examples of usage and workflow
   help                         : Show usage info
   ip                           : IPv4 related info
   ipv6                         : IPv6 related info
   json                         : Provide output in JSON
   list                         : Show all help commands
   notifier                     : Start/Stop/Restart/Query status of netq notifier
   origin                       : Owner of route or mac
   show                         : Show fabric-wide info about specified object
   verbose                      : More help on help
   <example>                    : name of example
   <hostname>                   : Name of the remote node you want to query
   <ip-server>                  : IP address of DB server
   <ipv4>                       : IPv4 address (no mask)
   <ipv6>                       : IPv6 address (no mask)
   <ip>                         : IPv4 or v6 address (no mask)
   <ipv4/prefixlen>             : IPv4 address with mask
   <ipv6/prefixlen>             : IPv6 address with mask
   <ip/prefixlen>               : IPv4 or IPv6 address with mask
   <mac>                        : MAC address
   <role>                       : Docker Swarm Node Role (worker, master)
   <text-time>                  : Time string such as 10m, 30s, 1h
   <text-endtime>               : Time string such as 10m, 30s, 1h
   <vrf>                        : Backend VRF
   <text-vlan>                  : VLAN

"""
from __future__ import absolute_import
from __future__ import print_function

from collections import OrderedDict
import functools
import ujson as json
import os
import sys
import signal
import time
from datetime import datetime
import re
import imp as importlib
import inspect
from functools import wraps
# Don't remove this import, used below
from operator import itemgetter  # noqa

from redis.sentinel import MasterNotFoundError

import tabulate
from network_docopt import (
    NetworkDocopt,
)

from netq_lib.common import utils
from netq_lib.common.enums import (
    BgColors,
    DiffState,
)
from netq_apps.lib.netq import (
    NetQ,
)
from netq_apps.modules.core.netq.config import (
    conf_get_experimental,
    conf_get_addons
)

from netq_apps.modules.core.netq.backend.config_ha_impl import (
    is_ts_server,
    ensure_environ,
)
import netq_apps.lib.completers as completers
import netq_apps
from netq_apps.lib.query import GatewayQuery

_PROC_NAME = 'netq-cli'
args = None
RC_SUCCESS = 0
RC_CMD_NOT_FOUND = -2
MAX_RESULTS_PER_SHOW = 5000

class Context(object):
    '''Allow passage of some key params between lower function
    and decorator'''
    def __init__(self, cli):
        self.col_headers = []
        self.col_sizes = []
        self.sort_cols = None
        self.sort_fn_str = None
        self.reverse = False
        self.changes = True if cli.get('changes') else False
        self.count = True if cli.get('count') else False


class MyTabulate(object):
    @staticmethod
    def __line_row(colsizes, *_):
        return ' '.join('-' * colsize for colsize in colsizes)

    @classmethod
    def __build_row(cls, colsizes, sep, cell_values, _, colaligns):
        values_with_attrs = []
        new_cell_values = []
        for cell_value, colsize, colalign in zip(cell_values,
                                                 colsizes,
                                                 colaligns):
            color = None
            if colalign == 'right':
                fmt = '{0:>%d}' % colsize
            else:
                fmt = '{0:<%d}' % colsize
            key = cell_value.strip()
            if key.startswith(BgColors.RED):
                color = BgColors.RED
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.YELLOW):
                color = BgColors.YELLOW
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.GREEN):
                color = BgColors.GREEN
                key = fmt.format(utils.strip_colors(key).strip())
            elif key.startswith(BgColors.GRAY):
                color = BgColors.GRAY
                key = fmt.format(utils.strip_colors(key).strip())
            else:
                key = fmt.format(key)
            if len(key) > colsize:
                idx = key.rfind(sep, 0, colsize)
                if idx == -1:
                    idx = colsize - 1
                if color:
                    new_cell_values.append(utils.color_wrap(color,
                                                            key[idx + 1:]))
                else:
                    new_cell_values.append(key[idx + 1:])
                key = key[:idx + 1] + ' ' * (colsize - idx - 1)
            else:
                new_cell_values.append('')
            if color:
                values_with_attrs.append(utils.color_wrap(color, key))
            else:
                values_with_attrs.append(key)
        line = ' '.join(values_with_attrs).rstrip()
        if any(new_cell_values):
            return line + '\n' + cls.__build_row(colsizes, sep,
                                                 new_cell_values, _,
                                                 colaligns)
        else:
            return line

    @classmethod
    def paginate(cls, data, headers, colsizes, sep=', ', first_entry=True):
        linebelowheader = functools.partial(cls.__line_row, colsizes)
        headerrow = datarow = functools.partial(cls.__build_row, colsizes, sep)
        simple = tabulate.TableFormat(
            lineabove=None, padding=0, linebelowheader=linebelowheader,
            linebetweenrows=None, linebelow=None, headerrow=headerrow,
            datarow=datarow, with_header_hide=['lineabove', 'linebelow'])
        if first_entry:
            print(tabulate.tabulate(data, headers=headers, tablefmt=simple))
        else:
            print(tabulate.tabulate(data, tablefmt=simple))


def _build_diff(entries, state, mods):
    output = []
    for entry in entries:
        if entry is None:
            state += 1
            continue
        state_str = DiffState.MAP[state]
        if state == DiffState.MODIFIED:
            mods.append(entry)
            if len(mods) == 2:
                if mods[0][:-1] != mods[1][:-1]:
                    output.append((state_str,) + tuple(mods[0]))
                    output.append(('',) + tuple(mods[1]))
                mods[:] = []
        else:
            output.append((state_str,) + tuple(entry))
    return output, state


def netq_tabulate(output, headers=None, jsonify=False, first=False,
                  first_entry=True, colsizes=None, **kwargs):
    if jsonify:
        out = []
        sanitized_headers = []
        json_out = []
        for header in headers:
            if header.startswith('#'):
                header = re.sub(r'#', 'num', header)
            sanitized_headers.append(
                ''.join(i.lower() if idx == 0 else i.lower().title()
                        for idx, i in enumerate(header.split(' ')))
            )
        for ele in output:
            out.append(dict((key, utils.strip_colors(item))
                            for key, item in zip(sanitized_headers, ele)))
        json_out.append(out)
        return json_out[0]
    else:
        # Stringify timestamps
        new_output = []
        for row in output:
            new_ele = []
            for ele in row:
                if isinstance(ele, float):
                    # If this is just a random float, don't convert.
                    # We assume random float if the number is less
                    # than 10 million (just a random pick)
                    if ele > 10000000:
                        try:
                            datetime.fromtimestamp(ele)
                            new_ele.append(utils.pretty_date(ele))
                            continue
                        except Exception:  # pylint: disable=broad-except
                            pass
                new_ele.append(ele)
            new_output.append(new_ele)
        MyTabulate.paginate(new_output, headers=headers, colsizes=colsizes,
                            first_entry=first_entry)

# These cli_show* routines never print the JSON output. This allows the
# JSON output to be aggregated. This is done already with 'show changes'
# command


class NetqCLI(NetworkDocopt):

    def __init__(self, daemon, logger, auth):

        ''''''
        global __doc__
        self.auth = auth
        GatewayQuery.set_auth(self.auth)
        self.callbacks = {}
        self.completers = []
        self.changes_cb = {}
        self._logger = logger

        # Make sure we update __doc__ as its used in help display
        self.cmdstrings = self.add_cmds(__doc__)
        NetworkDocopt.__init__(self, self.cmdstrings, daemon)

        for completer in self.completers:
            completer(self)

        # Add custom completers
        self.add_tag_matcher('<hostname>', completers.hostname_match, True)
        self.add_tab_complete_hook('<hostname>', completers.hostname_complete)

        self.short_doc_string = '''
    netq command is in one the of the following forms:
       netq config (config commands)
       netq [<hostname>] (check|show) <object> <options>
       netq trace <options>
       netq resolve
       netq (agent|notifier) (<start|stop|restart|status>) [vrf <vrf>]

    Type:
       'netq show agents' to see the list of hosts monitored by netq
       'netq help verbose' to get the help on help
       'netq help list' to dump all the commands and keyword help
       'netq example' to see examples of usage and possible workflow
       'netq help <text-keywords>' ('netq help check bgp', for eg.) to see
       commands matching keywords'''
        self.verbose_doc_string = '''
       netq commands have the following format:
           netq [<hostname>] action object [options]

       [] denotes an optional parameter or keyword
       <> denotes a parameter to be specified such as an IP prefix, addr etc.

       Hitting the TAB key will automatically show the available options.
       Partial keywords are also accepted; e.g.: 'netq show ip ro 4.0.0.1'.

       action is one of:
           check       :  Perform fabric-wide checks
           config      :  Configure NetQ
           example     :  Show examples of usage and workflow
           help        :  Show usage info
           resolve     :  Annotate input with names and interesting info
           show        :  Show fabric-wide info about specified object
           trace       :  Control plane trace path across fabric

       object is what the information is requested about such as interfaces,
       routes, services, etc.

       Options are used to further qualify the output desired such as
       asking for JSON output or requesting information about a specific object
       within the main object.

       Time based displays are governed by two options:
           around            : Go back in time to around ...
           between           : specify start and end times for output desired

       Time itself is specified as either a relative time such as 1s, 10m, 1d
       etc. or precise time '%Y-%m-%dT%H:%M:%S.%fZ' or 'd:h:m:s'.
        '''

    def add_cmds(self, doc):
        '''Construct complete command set and return for docopt to process'''

        options_idx = doc.find('Options:')
        usage = doc[:options_idx]
        post_usage = doc[options_idx:]

        mod_usage, mod_options, mod_callbacks = self._add_module_cmds()

        # new_usage = (usage + mod_usage + dyn_usage + post_usage +
        #             mod_options + dyn_options)

        new_usage = usage + mod_usage + post_usage + mod_options

        return new_usage

    @ensure_environ
    def _add_module_cmds(self):
        options = ''
        usage = ''
        callbacks = {}
        main_dir = os.path.dirname(inspect.getfile(netq_apps))
        python_version = '%d.%d' % (sys.version_info.major,
                                    sys.version_info.minor)
        local_dir = '/usr/local/lib/%s/dist_packages/netq_apps/modules/' % (
            python_version)

        # Always parse in this order: core, experimental, addons under main
        # python dist-packages dir, followed by experimental and addons under
        # local lib
        dirstocheck = [main_dir + '/modules/core']
        load_exp = conf_get_experimental()
        load_addons = conf_get_addons()
        if load_exp:
            dirstocheck.append(main_dir + '/modules/experimental')
        if load_addons:
            dirstocheck.append(main_dir + '/modules/addons')

        # Now for the local dirstocheck
        dirstocheck.append(local_dir + '/modules/core')
        if load_exp:
            dirstocheck.append(local_dir + '/modules/experimental')
        if load_addons:
            dirstocheck.append(local_dir + '/modules/addons')

        for moddir in dirstocheck:
            if not os.path.exists(moddir):
                continue
            # traverse root directory, and list directories as dirs and
            # files as files
            for root, dirs, files in os.walk(moddir):
                for filename in sorted(files):
                    # skip loading modules in the ts directory
                    if not is_ts_server() and 'ts' in root:
                        continue

                    modname, extn = os.path.splitext(
                        os.path.basename(filename))
                    mname = root + '/' + filename

                    try:
                        if extn == '.so':
                            mod = importlib.load_dynamic(modname, mname)
                        elif extn == '.py' and modname != '__init__':
                            mod = importlib.load_source(modname, mname)
                        else:
                            continue
                    except ImportError as ex:
                        self._logger.error('Unable to import module %s.'
                                           'Error is %s' % (mname, str(ex)))
                        continue
                    except Exception as ex:
                        self._logger.error('Unable to import module %s.'
                                           'Error is %s' % (mname, str(ex)))
                        continue
                    if hasattr(mod, 'build_doc_string'):
                        doc = mod.build_doc_string()
                    else:
                        doc = mod.__doc__
                    if not doc:
                        continue
                    usage_idx = doc.find('Usage:') + len('Usage:') + 1
                    options_idx = doc.find('Options:')
                    if usage_idx:
                        usage += doc[usage_idx:options_idx]
                    if options_idx:
                        options_idx += len('Options:') + 1
                        options += doc[options_idx:]
                    if (not hasattr(mod, 'app') or
                            not hasattr(mod.app, 'routes')):
                        continue
                    routes_from = mod.app.routes
                    if not isinstance(routes_from, dict):
                        self._logger.error(
                            'Unable to import module %s. Routes is not dict'
                            % (mname))
                        continue

                    for key, val in routes_from.iteritems():
                        if key in self.callbacks:
                            self._logger.warning(
                                'Duplicate route %s in module %s. '
                                'Prev route from %s' % (
                                    key, modname,
                                    self.callbacks[key]['module']))
                            continue

                        self.callbacks[key] = {'module': modname, 'cb': val}

                    # Add custom completers/matchers, if any
                    if (hasattr(mod, 'add_completers') and
                            callable(mod.add_completers)):
                        self.completers.append(mod.add_completers)

                    if (hasattr(mod, 'get_changes') and
                            callable(mod.get_changes)):
                        self.changes_cb[modname] = mod.get_changes

                    self._logger.info('Successfully imported module %s from %s'
                                      % (modname, mname))
        return usage, options, callbacks

    def process_cmd(self, cli, netq, connection):
        # TBD: THIS NEEDS WORK. DON't CHECK IN THIS CODE LIKE THIS
        if self.args.get('<hostname>'):
            start_arg = 2
        else:
            start_arg = 1
        end_arg = start_arg + 4

        cmd_match = False
        jsonify = cli.get('json')
        status = 0
        for x in range(end_arg, start_arg, -1):
            cbstr = ' '.join(str(a) for a in self.argv_expanded[start_arg:x])
            data = self.callbacks.get(cbstr)
            if data:
                cmd_match = True
                cli['module_name'] = data['module']
                if (self.argv_expanded[start_arg] != 'show'):
                    # Only show command outputs are streamed, and if
                    # they're not show count commands
                    status = data['cb'](cli, netq)
                    if status is None:
                        # Due to show commands yielding, they cannot return
                        # status
                        status = RC_SUCCESS
                else:
                    if cbstr == 'show changes':
                        outgen = data['cb'](cli, netq, self.changes_cb)
                    # Show changes is special because it needs additional param
                    else:
                        outgen = data['cb'](cli, netq)

                    if outgen is None:
                        return status

                    if isinstance(outgen, int):
                        return outgen

                    # Streaming output
                    for ele in outgen:
                        if not isinstance(ele, int):
                            out = json.dumps({'data': ele})
                            connection.send(out + '\r')
                        else:
                            status = ele
                            return status
            if cmd_match:
                break

        if not cmd_match:
            if not jsonify:
                print('Unable to find command %s' % ' '.join(self.argv))
            else:
                json.dumps({'error': 'Unable to find command %s' %
                            ' '.join(self.argv)}, escape_forward_slashes=False)
            return -1

        if cli.get('parser_reinit', False):
            os.kill(os.getppid(), signal.SIGHUP)

        return status

    def run(self, argv, uid, gid, connection):
        """
        'parent' is the NetDaemon object used for checking user credentials.
        """
        global __doc__

        len_argv = len(argv)
        netq = None

        # Handle help separately
        # Special check for handling "netq help verbose" and "netq help legend"
        if len_argv == 1:
            print(self.short_doc_string)
            return 0
        elif argv[1] == 'help' or argv[-1] == 'help':
            if len_argv == 2:
                print(self.short_doc_string)
                return 0
            if argv[2] == 'verbose':
                print(self.verbose_doc_string)
                return 0
            if argv[2] == 'list':
                print(self.cmdstrings)
                return 0
            else:
                print_section = False
                section = ''
                for line in self.cmdstrings.splitlines():
                    if not line:
                        continue
                    if line == 'Usage:':
                        section = 'Commands:'
                        print_section = True
                    elif line == 'Options:':
                        section = 'Keywords:'
                        print_section = True
                    prev_match = True
                    if argv[-1] == 'help':
                        for ele in argv[1:-1]:
                            if not re.search(r'\b%s\b' % ele, line):
                                prev_match = False
                                break
                    else:
                        for ele in argv[2:]:
                            if not re.search(r'\b%s\b' % ele, line):
                                prev_match = False
                                break
                    if prev_match:
                        if print_section:
                            print(section)
                            print_section = False
                        print(line)
            return 0
        elif argv[1] == 'config':
            ret = NetworkDocopt.run(self, argv)  # Populates self.args
            # Older version of network docopt used to print in run.
            # Newer version returns a tuple with the second element as the
            # string to be printed.
            if ret and len(ret) > 1 and ret[1]:
                print("Error: " + ret[1])
            return self.process_cmd(self.args, netq, connection)

        ret = NetworkDocopt.run(self, argv)  # Populates self.args
        if ret and len(ret) > 1 and ret[1]:
            print("Error: " + ret[1])
        self.argv = argv

        start_time = end_time = None
        if self.args.get('around') or self.args.get('between'):
            start_time = self.args.get('<text-time>')
        if self.args.get('and'):
            end_time = self.args.get('<text-endtime>')
        if self.args.get('changes') and not end_time:
            start_time = 'now'
            end_time = '1h'

        try:
            netq = NetQ(None, start_time, end_time, self._logger)
        except MasterNotFoundError:
            raise ValueError('Cannot connect to backend server.')

        return self.process_cmd(self.args, netq, connection)


def netq_show(show_func):
    '''Decorator to wrap netq show functions'''

    @wraps(show_func)
    def __wrapper(cli, netq):
        context = Context(cli)
        first = True
        sort_cols = None
        reverse = False
        output = []

        jsonify = True if cli.get('json') else False
        module = cli.get('module_name', 'results')

        if context.count:
            out = show_func(cli, netq, context)
            if not (isinstance(out, int) or
                    isinstance(out, float) or
                    isinstance(out, long)):
                out = 0
            if not jsonify:
                print('Count of matching {} records: {}'.format(module, out))
            else:
                print(json.dumps({'count': out}))

            yield RC_SUCCESS
            return

        paginate = 100
        for entry in show_func(cli, netq, context):
            if first:
                # Build the column headers if not specified
                if not context.col_headers:
                    context.col_headers = [e for e in entry.keys()
                                           if e != 'error']

                if context.sort_cols:
                    # Don't check if this is evaluatable as we maybe
                    # compromised. We're running this as root and don't want
                    # to evaluate malicious code, introduced via say an
                    # addons module.
                    sort_cols = context.sort_cols
                    reverse = context.reverse

                if context.sort_fn_str:
                    sort_fn = context.sort_fn_str
                else:
                    sort_fn = None

                if not sort_cols:
                    if context.changes:
                        sort_cols = [-1, 0, 1]
                        reverse = True
                    else:
                        sort_cols = [0, 1, 2]

                yield({'col_headers': context.col_headers,
                       'col_sizes': context.col_sizes,
                       'sort_cols': sort_cols,
                       'function': show_func.__name__,
                       'sort_fn_str': sort_fn,
                       'reverse': reverse,
                       'jsonify': jsonify,
                       'module': module})
                first = False

            vals = process_show_entry(netq, entry)
            if len(vals):
                if not jsonify and not context.changes:
                    #
                    # Look for elements that are timestamp. 
                    #
                    now = time.mktime(datetime.utcnow().timetuple())
                    start_time = netq.start_time
                    if start_time:
                        time_offset = now - start_time
                    else:
                        time_offset = 0
                    for key in vals:
                        if isinstance(vals[key], float) and vals[key] > 10000000:
                            # If this is just a random float, don't convert.
                            # We assume random float if the number is less
                            # than 10 million (just a random pick)
                            try:
                                if key == 'Last Changed':
                                    vals[key] = datetime.fromtimestamp(vals[key]).ctime()
                                else:
                                    # Adjust the time offset relative to start_time
                                    # The last change field for non-json output
                                    # should reflect the change time since
                                    # the start_time as its reference point
                                    datetime.fromtimestamp(vals[key])
                                    vals[key] = vals[key] + time_offset
                            except Exception:
                                pass
                output.append(vals.values())

            if not paginate:
                yield(output)
                paginate = 100
                output = []

            paginate -= 1

        if output:
            yield(output)
        elif not output and first:
            yield({'col_headers': context.col_headers,
                   'col_sizes': context.col_sizes,
                   'sort_cols': [],
                   'sort_fn_str': '',
                   'function': show_func.__name__,
                   'reverse': reverse,
                   'jsonify': jsonify,
                   'module': module})
        return

    return __wrapper


def display_json(disp_module, json_out, first_entry, last_entry):
    if first_entry and last_entry:
        print(json.dumps({disp_module: json_out,
                          'truncatedResult': False}, indent=4,
                         escape_forward_slashes=False))
    elif first_entry and not last_entry:
        dump = json.dumps({disp_module: json_out}, indent=4,
                          escape_forward_slashes=False)
        lines = dump.splitlines()
        print('\n'.join(lines[:-3]))
    elif not first_entry and not last_entry:
        dump = json.dumps({disp_module: json_out}, indent=4,
                          escape_forward_slashes=False)
        lines = dump.splitlines()
        print('\t},\n' + '\n'.join(lines[2:-3]))
    elif not first_entry and last_entry:
        print('\t}\n    ],\n    "truncatedResult":false\n}')


def display_stream(output, first_entry=False, last_entry=False):
    if not output:
        print('Query returned no output result')
        return

    if len(output) and len(output[0]) == 0:  # output == [{}]
        return

    metadata = output[0]
    sort_cols = metadata.get('sort_cols', [0, 1, 2])
    sort_fn_str = metadata.get('sort_fn_str', 'itemgetter')
    reverse = metadata.get('reverse', False)
    col_headers = metadata.get('col_headers', [])
    col_sizes = metadata.get('col_sizes', [])
    jsonify = metadata.get('jsonify', False)
    disp_module = metadata.get('module', 'records')

    if sort_fn_str is None:
        sort_fn_str = 'itemgetter'
    output.pop(0)
    if jsonify:
        json_out = netq_tabulate(
                       output, headers=col_headers, colsizes=col_sizes,
                       first_entry=first_entry, jsonify=jsonify)
        display_json(disp_module, json_out, first_entry, last_entry)
    else:
        if not output:
            if first_entry:
                print('\nNo matching {} records found'.format(disp_module))
        else:
            if first_entry:
                print('\nMatching {} records:'.format(disp_module))
            sort_key = eval(sort_fn_str)(*sort_cols)
            output.sort(key=sort_key, reverse=reverse)
            try:
                netq_tabulate(
                    output, headers=col_headers, colsizes=col_sizes,
                    first_entry=first_entry, jsonify=jsonify)
            except (IOError, KeyboardInterrupt):
                pass


def process_show_entry(netq, entry):
    '''Process a single entry and return the list'''
    vals = OrderedDict()
    if 'error' in entry:
        color = BgColors.RED
        entry.pop('error')
    elif (('Hostname' in entry and
           entry['Hostname'] in netq.rotten_nodes) or
          ('hostname' in entry and
           entry['hostname'] in netq.rotten_nodes) or
          ('Master' in entry and
           entry['Master'].split(':')[0] in netq.rotten_nodes)):
        color = BgColors.GRAY
    else:
        color = BgColors.ENDC

    for key in entry:
        if isinstance(entry[key], str):
            vals[key] = utils.color_wrap(color, entry[key])
        else:
            vals[key] = entry[key]

    return vals
