# Copyright 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from confluent_kafka import KafkaError
from confluent_kafka.avro import AvroConsumer
from confluent_kafka.avro.serializer import SerializerError
from json import dumps as json_dumps, loads as json_loads
from netq_lib.common import configuration
from netq_apps.lib.netq import NetQ
from netq_apps.cmd.netq import NetqCLI
from netq_apps.modules.core.trace.trace import cli_trace_path
from netq_apps.modules.core.os.ntp import cli_check_ntp
from netq_apps.modules.core.l3.bgp import cli_check_bgp
from netq_apps.modules.core.l3.mtu import cli_check_mtu
from netq_apps.modules.core.l3.ospf import cli_check_ospf
from netq_apps.modules.core.nvo.evpn import cli_check_evpn
from netq_apps.modules.core.l2.clag import cli_check_clag
from netq_apps.modules.core.l2.vlan import cli_check_vlan
from netq_apps.modules.core.platform.cables import cli_check_physical_links
from netq_apps.modules.core.platform.inventory import cli_check_license
from netq_apps.modules.core.platform.sensors import cli_check_sensors
from netq_apps.modules.core.netq.agents import cli_check_agents
from netq_apps.modules.core.nvo.lnv import cli_check_lnv
from netq_apps.modules.core.nvo.vxlan import cli_check_vxlan
from netq_apps.lib.auth import GatewayAuth
import sys
import os
from cStringIO import StringIO

from confluent_kafka import avro
from confluent_kafka.avro import AvroProducer
import multiprocessing
import time
import netq_lib.orm.avro.tc_response as tc_response
from netq_lib.common import utils
from network_docopt import NetworkDocopt

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend("".join(self._stringio.getvalue()))
        sys.stdout = self._stdout

def kafka_consume(conf_dict):
    logger = utils.get_logger("netq_kafka_bridge", conf_dict)
    logger.info(conf_dict)
    bootstrap_server = os.environ['KAFKA_BOOTSTRAP_SERVERS']
    schema_registry_url = "http://" + os.environ['SCHEMA_REGISTRY_URL']
    if ('TC_INPUT_TOPIC' in os.environ):
        input_topic = os.environ['TC_INPUT_TOPIC']
    else:
        input_topic = "netq_tc_request"
    if ('TC_OUTPUT_TOPIC' in os.environ):
        output_topic = os.environ['TC_OUTPUT_TOPIC']
    else:
        output_topic = "netq_tc_response"
    c = AvroConsumer({
        'bootstrap.servers': bootstrap_server,
        'group.id': 'group-1',
        'schema.registry.url': schema_registry_url})

    c.subscribe([input_topic])
    if (('netq-cli' in conf_dict) and ('server' in conf_dict['netq-cli'])):
        gateway_ip = conf_dict['netq-cli']['server']
    else:
        gateway_ip = os.environ['NETQ_API_GATEWAY_SERVICE_HOST']
    if (('netq-cli' in conf_dict) and ('port' in conf_dict['netq-cli'])):
        gateway_port = conf_dict['netq-cli']['port']
    else:
        gateway_port = os.environ['NETQ_API_GATEWAY_SERVICE_PORT']

    while True:
        try:
            msg = c.poll(10)
        except SerializerError as e:
            print("Message deserialization failed for {}: {}".format(msg, e))
            break

        if (msg is None):
            continue

        if msg.error():
            if msg.error().code() == KafkaError._PARTITION_EOF:
                continue
            else:
                print(msg.error())
                break

        if (msg.value()['params'] is None):
            continue

        auth = GatewayAuth(logger=logger,
                           gateway_ip=gateway_ip,
                           gateway_port=gateway_port)
        cli = NetqCLI(False, logger, auth)
        params = json_loads(msg.value()['params'])
        msg_type = msg.value()['request_type']
        if (msg_type == 0):
            reqJson = ["netq", "trace", params['dst']]
            if (int(params['vlan']) > 0):
                reqJson.extend(["vlan", params['vlan']])
            reqJson.extend(["from", params['src'], "json"])
        elif (msg_type == 1):
            reqJson = ["netq", "check", params['validation_type'], "json"]
        NetworkDocopt.run(cli, reqJson)
        reqJson = cli.args
        netq = NetQ(None, None, None, None)
        rc = 0
        print(reqJson)

        try:
            with Capturing() as resJson:
                if (msg_type == 0):
                    rc = cli_trace_path(reqJson, netq)
                elif (msg_type == 1):
                    validation_type = params['validation_type']
                    if (validation_type == "bgp"):
                        rc = cli_check_bgp(reqJson, netq)
                    elif (validation_type == "evpn"):
                        rc = cli_check_evpn(reqJson, netq)
                    elif (validation_type == "clag"):
                        rc = cli_check_clag(reqJson, netq)
                    elif (validation_type == "interfaces"):
                        rc = cli_check_physical_links(reqJson, netq)
                    elif (validation_type == "license"):
                        rc = cli_check_license(reqJson, netq)
                    elif (validation_type == "agents"):
                        rc = cli_check_agents(reqJson, netq)
                    elif (validation_type == "ntp"):
                        rc = cli_check_ntp(reqJson, netq)
                    elif (validation_type == "sensor"):
                        rc = cli_check_sensors(reqJson, netq)
                    elif (validation_type == "lnv"):
                        rc = cli_check_lnv(reqJson, netq)
                    elif (validation_type == "vxlan"):
                        rc = cli_check_vxlan(reqJson, netq)
                    elif (validation_type == "vlan"):
                        rc = cli_check_vlan(reqJson, netq)
                    elif (validation_type == "mtu"):
                        rc = cli_check_mtu(reqJson, netq)
                    elif (validation_type == "ospf"):
                        rc = cli_check_ospf(reqJson, netq)
            job_status = "Complete"
        except Exception:
            rc = 1
            reqJson = ""
            job_status = "Complete with exceptions"

        respStr = ""
        respStr += ' '.join(json_dumps("".join(list(resJson))).decode('string-escape').strip('"').split())

        job_details = {"opid": msg.value()['opid'], "jobid": msg.value()['jobid'], "job_req_time": msg.value()['job_req_time'],
                       "job_start_time": msg.value()['job_start_time'], "job_end_time": long(time.time() * 1000),
                       "job_status": job_status, "status_details": str(rc), "is_scheduled": msg.value()['is_scheduled'],
                       "request_type": msg.value()['request_type'], "user_name": msg.value()['user_name'], "params": msg.value()['params']}
        value_schema = avro.loads(tc_response.value)
        key_schema = avro.loads(tc_response.key)
        value = {"job_request": job_details, "response": respStr}
        key = {"opid": msg.value()['opid']}

        avroProducer = AvroProducer({
          'bootstrap.servers': bootstrap_server,
          'schema.registry.url': schema_registry_url
        }, default_key_schema=key_schema, default_value_schema=value_schema)
        avroProducer.produce(topic=output_topic, value=value, key=key)
        avroProducer.flush()
        msg = None

        print(value, key)

    c.close()


if __name__ == "__main__":
    procs = 1
    jobs = []
    conf_dict = configuration.get_config("netq-cli")
    for i in range(0, procs):
        process = multiprocessing.Process(target=kafka_consume,
                args=(conf_dict,))
        jobs.append(process)

    for j in jobs:
        j.start()


    for j in jobs:
        j.join()

    print "Exiting TraceScheduler"
